#include <pic.h>

#include "main.h"
#include "commands.h"
#include "helper.h"
#include "serial.h"
#include "delay/delay.h"
#include "delay/delay_alternative_enhanced_precision.h"

extern enum { idle, rotatingcw, rotatingccw } status;

extern char out[STRINGLENGTH];
extern unsigned int pos;
extern unsigned int destpos;
extern char livepos;
extern char nocheck;
extern char destpos_use;
bank1 extern unsigned int speed;
bank1 extern unsigned int previouspos;
bank1 extern char timer2softscaler;

// prints current rotate pos
void putpos( void ) {
	unsigned int distance;

	int2str( pos, out, STRINGLENGTH );
	puts( "POS " );	puts( out );
	if( !nocheck ) {
		if( checkbounds( pos, WITHJITTER ) < 0 )
			puts( " lower than min bound" );
		if( checkbounds( pos, WITHJITTER ) > 0 )
			puts( " higher than max bound" );
	}

	if( status != idle ) {
		// displaying rotate speed / sec
		puts( " SPD " );
		int2str( speed, out, STRINGLENGTH ); puts( out );
	}

	if( livepos ) {
		if( destpos_use ) {
			puts( " DST " );
			int2str( destpos, out, STRINGLENGTH ); puts( out );

			if( destpos > pos )
				distance = destpos-previouspos;
			else
				distance = previouspos-destpos;

			puts( " EST " );
			int2str( distance / speed, out, STRINGLENGTH ); puts( out );
		}
		puts( "                                    " );
	}
	putch( 13 );
}

void starttimer2( void ) {
	previouspos = pos;
	timer2softscaler = 0;
	TMR2ON = 1;
}

void stoptimer2( void ) {
	TMR2ON = 0; // timer 2 off
	speed = 0;
}

// don't call this directly, use STOPROTATE( char ) instead!
void stoprotate( char silent ) {
	RB1 = 0;
	RB2 = 0;
	RB3 = 0;
	RB4 = 0;
	status = idle;
	if( !silent )
		puts( "ROT rotation stopped\r\nSTS idle\r\n" );
	stoptimer2();
}	

// clockwise rotate - pot value decreases
// don't call this directly, use STARTROTATE_CW() instead!
void startrotate_cw( void ) {
	stoprotate( SILENT );
	RB1 = 0;
	RB2 = 0;
	RB3 = 1;
	RB4 = 1;
	starttimer2(); // timer 2 (pot idle checker) on
	status = rotatingcw;
	puts( "ROT cw rotation started\r\nSTS rotatingcw\r\n" );
}	

// counterclockwise rotate - pot value increases
// don't call this directly, use STARTROTATE_CCW() instead!
void startrotate_ccw( void ) {
	stoprotate( SILENT );
	RB1 = 1;
	RB2 = 1;
	RB3 = 0;
	RB4 = 0;
	starttimer2(); // timer 2 (pot idle checker) on
	status = rotatingccw;
	puts( "ROT ccw rotation started\r\nSTS rotatingccw\r\n" );
}	

void printversion( void ) {
	puts( "VER Antenna Rotator Controller v" VERSION "\r\n" );
	puts( "VER Built on " __DATE__ " at " __TIME__ "\r\n" );
	puts( "VER (C) by Nonoo (www.nonoo.hu)\r\n" );
}

void printhelp( void ) {
	puts( "HLP available commands: (these are not case sensitive)\r\n" );
	puts( "HLP H  - this help\r\n" );
	puts( "HLP S  - status\r\n" );
	puts( "HLP R0 - stop rotating\r\n" );
	puts( "HLP R1 - rotate clockwise - pos decreases\r\n" );
	puts( "HLP R2 - rotate counterclockwise - pos increases\r\n" );
	puts( "HLP D[num|name] - rotate to the given position\r\n" );
	puts( "HLP G[num|name] - same as above\r\n" );
	puts( "HLP W[name] (pos) - writes current/given pos to EEPROM\r\n" );
	puts( "HLP E[name|banknum] - erases the given name/banknum in the EEPROM\r\n" );
	puts( "HLP O  - lists saved positions from EEPROM\r\n" );
	puts( "HLP L0 - switching live reporting of current position off\r\n" );
	puts( "HLP L1 - switching live reporting of current position on (the controller reports current position live)\r\n" );
	puts( "HLP C0 - switching position bounds checking off (DANGEROUS, WATCH OUT)\r\n" );
	puts( "HLP C1 - switching position bounds checking on\r\n" );
	puts( "HLP Q  - reset\r\n" );
	puts( "HLP V  - returns version information\r\n" );
}

// loads the given eprbank saved position into savedpos
// returns 0 if there is no saved position in the given eprbank
char loadpos( char eprbank, savedpos_t* savedpos ) {
	char hi, lo, i = 0;

	lo = EEPROM_READ( sizeof( savedpos_t ) * eprbank );
	hi = EEPROM_READ( sizeof( savedpos_t ) * eprbank + 1 );
	savedpos->pos = ((unsigned int)hi << 8) + lo;
	while( i < SAVEDPOSNAMELENGTH ) {
		savedpos->name[i] = EEPROM_READ( sizeof( savedpos_t ) * eprbank + 2 + i );
		i++;
	}

	if( savedpos->pos == 0xffff )
		return 0;

	return 1;
}	

// loads the saved position with the given name into savedpos and returns the bank
// num on success, else ERR_NOSUCHBANK
char loadposname( char* name, savedpos_t* savedpos ) {
	char i = 0;

	while( i < SAVEDPOSBANKCOUNT ) {
		if( loadpos( i, savedpos ) ) {
			if( strncmp( name, savedpos->name, SAVEDPOSNAMELENGTH ) == 0 ) {
				return i;
			}
		}
		i++;
	}
	return ERR_NOSUCHBANK;
}

// writes the given savedpos structure to the eprbankth EEPROM bank
void writesavedpos( savedpos_t savedpos, char eprbank ) {
	char i = 0;

	EEPROM_WRITE( sizeof( savedpos_t ) * eprbank, (char)((savedpos.pos << 8) >> 8) );
	EEPROM_WRITE( sizeof( savedpos_t ) * eprbank + 1, (char)(savedpos.pos >> 8) );
	while( i < SAVEDPOSNAMELENGTH ) {
		EEPROM_WRITE( sizeof( savedpos_t ) * eprbank + 2 + i, savedpos.name[i] );
		i++;
	}

	while( WR ) {} // waiting for EEPROM write to complete
}

// saves the given position and name to the first free EEPROM bank,
// or if there is a position saved with the given name, updates it's value
void savepos( unsigned int pos, char* name ) {
	savedpos_t savedpos;
	char eprbank = 0;

	upcase( name );

	// searching for the given name in saved banks
	eprbank = loadposname( name, &savedpos );

	if( eprbank == ERR_NOSUCHBANK ) { // no saved position found with the given name
		// searching for the first free eeprom bank
		eprbank = 0;
		while( eprbank < SAVEDPOSBANKCOUNT ) {
			if( loadpos( eprbank, &savedpos ) == 0 ) {
				// free position found
				break;
			}
			eprbank++;
		}
	}

	if( eprbank == SAVEDPOSBANKCOUNT ) {
		puts( "ERR no free banks available\r\n" );
		return;
	}

	strncpy( savedpos.name, name, SAVEDPOSNAMELENGTH );
	savedpos.pos = pos;

	// writing to eeprom
	writesavedpos( savedpos, eprbank );

	puts( "EPW saved position " );
	int2str( savedpos.pos, out, STRINGLENGTH ); puts( out );
	puts( " as " );
	puts( savedpos.name );
	puts( " in bank #" );
	int2str( eprbank, out, STRINGLENGTH ); puts( out );
	putch( 10 ); putch( 13 );
}

// erases the given banknum, or the bank with the given name
void erasepos( char* in ) {
	savedpos_t savedpos;
	char banknum;

	parseparam( in, out, 0 );

	// searching for the bank name
	upcase( out );
	banknum = loadposname( out, &savedpos );
	if( banknum == ERR_NOSUCHBANK ) {
		// if no bank found with that name, trying to convert the name to int
		banknum = str2int( out );
		if( banknum == ERR_NOSUCHBANK ) {
			puts( "ERR unknown parameter\r\n" );
			return;
		}
	}

	if( banknum >= SAVEDPOSBANKCOUNT ) {
		puts( "ERR no such bank\r\n" );
		return;
	}

	savedpos.pos = 0xffff;
	memset( savedpos.name, 0xff, SAVEDPOSNAMELENGTH );
	writesavedpos( savedpos, banknum );

	puts( "EPE bank #" );
	int2str( banknum, out, STRINGLENGTH ); puts( out );
	puts( " erased\r\n" );
}

void listsavedpositions( void ) {
	char eprbank = 0;
	savedpos_t savedpos;

	while( eprbank < SAVEDPOSBANKCOUNT ) {
		puts( "EPL bank #" );
		if( eprbank < 10 ) // for fixed with pos display
			putch( '0' );
		int2str( eprbank, out, STRINGLENGTH ); puts( out );
		puts( ": " );
		if( loadpos( eprbank, &savedpos ) ) {
			int2str( savedpos.pos, out, STRINGLENGTH ); puts( out );

			// for fixed with pos display
			if( savedpos.pos < 1000 )
				putch( ' ' );
			if( savedpos.pos < 100 )
				putch( ' ' );
			if( savedpos.pos < 10 )
				putch( ' ' );
			putch( ' ' );
			puts( savedpos.name );
		} else {
			// no data saved in bank
			puts( "empty" );
		}
		putch( 10 ); putch( 13 );
		eprbank++;
	}
}

// starts rotation to the previously set destpos value or
// if posname is given, loads it's value from EEPROM and starts rotation to it
void rotatetopos( char* in ) {
	savedpos_t savedpos;
	char eprbank;

	parseparam( in, out, 0 );

	// searching for the bank name
	upcase( out );
	eprbank = loadposname( out, &savedpos );
	if( eprbank == ERR_NOSUCHBANK ) {
		// if no bank found with that name, trying to convert the name to int
		destpos = str2int( out );
		if( destpos == ERR_INTCONV ) {
			puts( "ERR no such bank\r\n" );
			return;
		}
	} else {
		destpos = savedpos.pos;
		puts( "EPC destination position loaded from bank #" );
		int2str( eprbank, out, STRINGLENGTH ); puts( out );
		putch( 10 ); putch( 13 );
	}

	if( destpos-pos < MAXJITTER || pos-destpos < MAXJITTER ) {
		puts( "ERR rotator is in given position\r\n" );
		return;
	}

	if( !nocheck ) {
		if( checkbounds( destpos, WITHJITTER ) != 0 ) {
			puts( "ERR destination position is out of bounds\r\n" );
			return;
		}

		if( ( pos < destpos ) && ( checkbounds( pos, WITHOUTJITTER ) > 0 ) ) {
			// ccw rotation needed, but we're on the limit
			puts( "ERR max bound reached\r\n" );
			return;
		}
		if( ( pos > destpos ) && ( checkbounds( pos, WITHOUTJITTER ) < 0 ) ) {
			// cw rotation needed, but we're on the limit
			puts( "ERR min bound reached\r\n" );
			return;
		}
	}

	puts( "DST destination position is " );
	int2str( destpos, out, STRINGLENGTH ); puts( out );
	puts( ", current position: " );
	int2str( pos, out, STRINGLENGTH ); puts( out );

	destpos_use = 1;

	if( pos < destpos ) {
		puts( ", ccw rotation needed\r\n" );
		STARTROTATE_CCW();
	}
	else {
		puts( ", cw rotation needed\r\n" );
		STARTROTATE_CW();
	}
}
