#include <pic.h>

#include "main.h"
#include "serial.h"
#include "helper.h"
#include "commands.h"

__CONFIG( MCLREN & UNPROTECT & INTCLK & WDTEN );

enum { idle, rotatingcw, rotatingccw } status = idle;

bank1 char in[STRINGLENGTH];
bank1 char inp = 0; // input string length counter

bank1 char out[STRINGLENGTH];
unsigned int pos = 0xffff;
unsigned int destpos = 0xffff;
char a_derrcount = 0;
char livepos = 0;
char nocheck = 0;
char nocheckpot = 0; // timer2 pot check
char destpos_use = 0;

// timer1 stuff
char ledflashcount = 0;

// timer2 stuff - pot change protection
// if rotating is in progress, timer2 periodically checks if the pot
// value has been changed - if there's no change, it stops rotation
bank1 char timer2softscaler = 0;
bank1 unsigned int speed = 0;
bank1 unsigned int previouspos = 0xffff;
bank1 char stopflag = 0; // this is set to 1 if there's an error with the pot

enum { nobuttonpressed, cwbuttonpressed, ccwbuttonpressed } buttonstatus = nobuttonpressed;
char cwbuttonbounce = 0;
char ccwbuttonbounce = 0;

void checkcommand( char* in ) {
	unsigned int tmp;

	putch( 10 ); putch( 13 );
	switch( in[0] ) {
		case 's': // status report
		case 'S':
			switch( status ) {
				case idle: puts( "STS idle\r\n" ); break;
				case rotatingcw: puts( "STS rotatingcw\r\n" ); break;
				case rotatingccw: puts( "STS rotatingccw\r\n" ); break;
				default: puts( "ERR unknown state\r\n" ); break;
			}

			putpos(); putch( 10 );

			puts( "BND min bound: " );
			int2str( ROTATEMINVAL, out, STRINGLENGTH ); puts( out );
			puts( " max bound: " );
			int2str( ROTATEMAXVAL, out, STRINGLENGTH ); puts( out );
			putch( 10 ); putch( 13 );
			if( nocheck ) puts( "BCK position bounds checking off\r\n" );
			else puts( "BCK position bounds checking on\r\n" );
			if( nocheckpot ) puts( "PCK pot turn checking off\r\n" );
			else puts( "PCK pot turn checking on\r\n" );
			if( livepos ) puts( "LIV live position reporting on\r\n" );
			else puts( "LIV live position reporting off\r\n" );
			if( destpos_use ) {
				puts( "DST destination position is " );
				int2str( destpos, out, STRINGLENGTH ); puts( out );
				putch( 10 ); putch( 13 );

				if( destpos > pos )
					tmp = destpos-previouspos;
				else
					tmp = previouspos-destpos;

				puts( "EST " );
				int2str( tmp / speed, out, STRINGLENGTH ); puts( out );
				putch( 10 ); putch( 13 );
			} else
				puts( "DST no destination position\r\n" );

			break;

		case 'r': // rotate start/stop
		case 'R':
			switch( in[1] ) {
				case '1': // clockwise rotate
					if( !nocheck && checkbounds( pos, WITHJITTER ) < 0 ) {
						puts( "ERR min bound reached\r\n" );
						break;
					}
					STARTROTATE_CW();
					destpos_use = 0; // switching destination position off
					break;
				case '2': // counterclockwise rotate
					if( !nocheck && checkbounds( pos, WITHJITTER ) > 0 ) {
						puts( "ERR max bound reached\r\n" );
						break;
					}
					STARTROTATE_CCW();
					destpos_use = 0; // switching destination position off
					break;
				case '0':
					STOPROTATE( VERBOSE );
					destpos_use = 0; // switching destination position off
					break;
				default: puts( "ERR unknown parameter\r\n" );
			}
			break;

		case 'l': // live reporting
		case 'L':
			switch( in[1] ) {
				case '1': puts( "LIV live position reporting on\r\n" ); livepos = 1; break;
				case '0': puts( "LIV live position reporting off\r\n" ); livepos = 0; break;
				default: puts( "ERR unknown parameter\r\n" );
			}
			break;

		case 'c': // position bounds checking on/off
		case 'C':
			switch( in[1] ) {
				case '0': puts( "BCK position bounds checking off\r\n" ); nocheck = 1; break;
				case '1': puts( "BCK position bounds checking on\r\n" ); nocheck = 0; break;
				default: puts( "ERR unknown parameter\r\n" );
			}
			break;

		case 'p': // pot turn checking on/off
		case 'P':
			switch( in[1] ) {
				case '0': puts( "PCK pot turn checking off\r\n" ); nocheckpot = 1; break;
				case '1': puts( "PCK pot turn checking on\r\n" ); nocheckpot = 0; break;
				default: puts( "ERR unknown parameter\r\n" );
			}
			break;

		case 'g': // rotate to the given position
		case 'G':
		case 'd':
		case 'D':
			rotatetopos( &in[1] );
			break;

		case 'q': // reset
		case 'Q':
			puts( "RST soft-reset\r\n" );
			asm( "clrf _PCLATH" );
			asm( "goto 0" );
			break;

		case 'v': // version info
		case 'V':
			printversion();
			break;

		case 'w': // saving a position into EEPROM
		case 'W':
			parseparam( &in[1], out, 0 );
			if( *out == 0 ) {
				puts( "ERR unknown parameter\r\n" );
				break;
			}
			parseparam( &in[1], out, 1 );
			if( *out == 0 ) { // no second parameter, saving current pos
				parseparam( &in[1], out, 0 );
				savepos( pos, out );
			} else { // saving given pos
				tmp = str2int( out );
				if( tmp == ERR_INTCONV ) {
					puts( "ERR unknown parameter\r\n" );
					break;
				}
				parseparam( &in[1], out, 0 );
				savepos( tmp, out );
			}
			break;

		case 'e': // erases the given name/banknum in the EEPROM
		case 'E':
			erasepos( &in[1] );
			break;

		case 'o': // lists saved positions from EEPROM
		case 'O':
			listsavedpositions();
			break;

		case 'h': // online help
		case 'H':
			printhelp();
			break;

		default:
			puts( "ERR unknown command\r\n" );
	}
	inp = 0;
	memset( in, 0, STRINGLENGTH );
}	

void mainloop( void ) {
	// checking for a char in RX register
	if( RCIF ) {
		dummy = getch();
		putch( dummy );
		switch( dummy ) {
			case 13: // enter pressed
				in[inp] = 0;
				checkcommand( in );
				break;
			case 8: // backspace pressed
				if( inp > 0 )
					inp--;
				in[inp] = 0;
				break;
			default:
				in[inp++] = dummy;
				if( inp == STRINGLENGTH ) {
					checkcommand( in );
				}
		}
	} else {
		clearUSARTErrors;
	}

	// checking buttons
	if( CWBUTTON && status == idle ) {
		cwbuttonbounce++;
		if( cwbuttonbounce > 100 ) {
			cwbuttonbounce = 0;
			buttonstatus = cwbuttonpressed;
			STARTROTATE_CW();
		}
	}
	if( CCWBUTTON && status == idle ) {
		ccwbuttonbounce++;
		if( ccwbuttonbounce > 100 ) {
			ccwbuttonbounce = 0;
			buttonstatus = ccwbuttonpressed;
			STARTROTATE_CCW();
		}
	}
	if( buttonstatus != nobuttonpressed ) { // there was a button pressed last time
		if( !CWBUTTON && !CCWBUTTON ) { // if there's no button pressed currently
			cwbuttonbounce = ccwbuttonbounce = 0;
			buttonstatus = nobuttonpressed;
			STOPROTATE( VERBOSE );
		}
	}

	// checking for a/d result
	if( !GODONE ) {
		a_derrcount = 0;
		pos = ((unsigned int)ADRESH << 8) + ADRESL;

		GODONE = 1; // starting conversion
	} else {
		a_derrcount++;
	}

	// if there were no a/d results for a few times
	if( a_derrcount > MAXADERROR ) {
		a_derrcount = 0;
		puts( "ERR a/d timeout\r\n" );
		STOPROTATE( VERBOSE );
	}

	if( livepos ) { // live position reporting
		putpos();
	}

	// if a destination position has been given
	if( destpos_use ) {
		if( ( destpos <= pos && status == rotatingccw ) ||
		    ( destpos >= pos && status == rotatingcw ) ) {
			puts( "DST destination position " );
			int2str( destpos, out, STRINGLENGTH ); puts( out );
			puts( " reached\r\n" );
			STOPROTATE( VERBOSE );
			destpos_use = 0; // switching destination position off
		}
	}

	// checking for min/max rotate values
	if( !nocheck ) {
		if ( status == rotatingcw && checkbounds( pos, WITHOUTJITTER ) < 0 ) {
			puts( "ERR min bound reached\r\n" );
			STOPROTATE( VERBOSE );
		}
		if( status == rotatingccw && checkbounds( pos, WITHOUTJITTER ) > 0 ) {
			puts( "ERR max bound reached\r\n" );
			STOPROTATE( VERBOSE );
		}
	}

	if( stopflag && !nocheckpot ) {
		stopflag = 0;
		puts( "ERR pot did not turn but rotation is on\r\n" );
		STOPROTATE( VERBOSE );
	}

	CLRWDT();
}

void interrupt interrupthandler( void ) {
	// timer1 interrupt
	if( TMR1IF ) {
		TMR1ON = 0; // timer 1 off
		TMR1IF = 0; // clear the timer1 flag

		// fast led blinking on comm
		if( ledflashcount ) {
			set16( TMR1H, TMR1L, 62500 );
			ledflashcount--;
			STATUSLED = !STATUSLED;
		} else {
			if( status != idle ) { // if rotate in progress
				set16( TMR1H, TMR1L, 30000 );
				STATUSLED = !STATUSLED;
			} else {
				set16( TMR1H, TMR1L, 62500 );
				STATUSLED = 1;
			}
		}

		TMR1ON = 1; // timer 1 on
	}

	// timer2 interrupt
	if( TMR2IF ) {
		TMR2ON = 0; // timer 2 off
		TMR2IF = 0; // clear the timer2 flag

		if( status != idle ) {
			// we're using a software postscaler to slow down the timer
			// to approx. 1 sec
			if( timer2softscaler++ == 17 ) {
				if( previouspos > pos )
					speed = previouspos-pos;
				else
					speed = pos-previouspos;

				// if the pot doesn't turn
				if(	!nocheckpot && speed < MAXJITTER ) {
					stopflag = 1;
					return;
				}

				previouspos = pos;
				timer2softscaler = 0;
			}

			TMR2ON = 1; // timer 2 on
		}
	}
}

void main() {
	// initializing ports
	ANSEL = 0;
	ANSELH = 0;

	TRISSTATUSLED = 0; // STATUSLED -> output
	STATUSLED = 1;

	TRISB1 = 0; // motor pins -> outputs
	TRISB2 = 0;
	TRISB3 = 0;
	TRISB4 = 0;
	RB1 = 0;
	RB2 = 0;
	RB3 = 0;
	RB4 = 0;

	// a/d init
	TRISPOT = 1; // POT -> input
	ANSEL = 0;
	ANSELH = 0;
	ANSPOT = 1; // setting POT port as analog input

	ADCS0 = 1; // clock from internal osc (500 kHz max.)
	ADCS1 = 1;
	ADFM = 1; // right justified format
	VCFG0 = 0; // setting Vdd as positive voltage reference
	VCFG1 = 0; // setting Vss as negative voltage reference

	ADON = 1; // a/d module on
	GODONE = 1; // starting conversion

	// timer1 init
	TMR1H = 0;
	TMR1L = 0;
	T1CKPS0 = 1; // prescale = 1:8
	T1CKPS1 = 1;
	TMR1CS = 0; // clock source = Fosc/4
	T1OSCEN = 0; // RC oscillator disable

	TMR1IE = 1; // enable timer1 interrupts
	TMR1IF = 0; // clear timer1 interrupt flag
	TMR1ON = 1; // timer 1 on

	// timer2 init
	T2CKPS0 = 1; // 1:16 prescaler
	T2CKPS1 = 1;
	TOUTPS0 = 1; // 1:14 postscaler
	TOUTPS1 = 0;
	TOUTPS2 = 1;
	TOUTPS3 = 1;
	PR2 = 0xf7; // setting compare value

	TMR2IE = 1; // enable timer2 interrupts
	TMR2IF = 0; // clear timer2 interrupt flag

	PEIE = 1; // enable peripherial interrupts
	ei(); // enable interrupts

	// button watch init
	TRISCWBUTTON = 1; // manual rotate buttons -> inputs
	TRISCCWBUTTON = 1;
	CWBUTTON = 0;
	CCWBUTTON = 0;

	initserialcomm();

	puts( "\r\n\r\nRST controller init finished\r\n\r\n" );
	printversion();

loop:
	mainloop();
	goto loop;
}
